/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author Sebastian Bittrich <sebastian.bittrich@rcsb.org>
 */

import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { StructureRepresentationPresetProvider, PresetStructureRepresentations } from '../../mol-plugin-state/builder/structure/representation-preset';
import { MembraneOrientationProvider, MembraneOrientation } from './prop';
import { StateObject, StateObjectRef, StateObjectCell, StateTransformer, StateTransform, StateBuilder } from '../../mol-state';
import { Task } from '../../mol-task';
import { PluginBehavior } from '../../mol-plugin/behavior';
import { MembraneOrientationRepresentationProvider, MembraneOrientationParams, MembraneOrientationRepresentation } from './representation';
import { HydrophobicityColorThemeProvider } from '../../mol-theme/color/hydrophobicity';
import { PluginStateObject, PluginStateTransform } from '../../mol-plugin-state/objects';
import { PluginContext } from '../../mol-plugin/context';
import { DefaultQueryRuntimeTable } from '../../mol-script/runtime/query/compiler';
import { StructureSelectionQuery, StructureSelectionCategory } from '../../mol-plugin-state/helpers/structure-selection-query';
import { MolScriptBuilder as MS } from '../../mol-script/language/builder';
import { GenericRepresentationRef } from '../../mol-plugin-state/manager/structure/hierarchy-state';
import { Expression } from '../../mol-script/language/expression';
import { StateTransforms } from '../../mol-plugin-state/transforms';
import { Color } from '../../mol-util/color';
import { createStructureRepresentationParams } from '../../mol-plugin-state/helpers/structure-representation-params';
import { PluginUIContext } from '../../mol-plugin-ui/context';
import { StateObjectSelector } from "../../mol-state/object";
import { applyTransformations, createMembraneOrientation, getAtomGroupExpression } from './transformation';
import { PDBTMDescriptor, PDBTMRegion, PMS } from './types';

type StructureComponentType = StateObjectSelector<
        PMS,
        StateTransformer<StateObject<any, StateObject.Type<any>>,
        StateObject<any, StateObject.Type<any>>, any>
    > | undefined;

const Tag = MembraneOrientation.Tag;

export const TMDETMembraneOrientation = PluginBehavior.create<{ autoAttach: boolean }>({
    name: 'tmdet-membrane-orientation-prop',
    category: 'custom-props',
    display: {
        name: 'Membrane Orientation',
        description: 'Data calculated with TMDET algorithm.'
    },
    ctor: class extends PluginBehavior.Handler<{ autoAttach: boolean }> {
        private provider = MembraneOrientationProvider

        register(): void {
            DefaultQueryRuntimeTable.addCustomProp(this.provider.descriptor);

            this.ctx.customStructureProperties.register(this.provider, this.params.autoAttach);

            this.ctx.representation.structure.registry.add(MembraneOrientationRepresentationProvider);
            this.ctx.query.structure.registry.add(isTransmembrane);

            this.ctx.genericRepresentationControls.set(Tag.Representation, selection => {
                const refs: GenericRepresentationRef[] = [];
                selection.structures.forEach(structure => {
                    const memRepr = structure.genericRepresentations?.filter(r => r.cell.transform.transformer.id === MembraneOrientation3D.id)[0];
                    if (memRepr) refs.push(memRepr);
                });
                return [refs, 'Membrane Orientation'];
            });
            this.ctx.builders.structure.representation.registerPreset(MembraneOrientationPreset);
        }

        update(p: { autoAttach: boolean }) {
            let updated = this.params.autoAttach !== p.autoAttach;
            this.params.autoAttach = p.autoAttach;
            this.ctx.customStructureProperties.setDefaultAutoAttach(this.provider.descriptor.name, this.params.autoAttach);
            return updated;
        }

        unregister() {
            DefaultQueryRuntimeTable.removeCustomProp(this.provider.descriptor);

            this.ctx.customStructureProperties.unregister(this.provider.descriptor.name);

            this.ctx.representation.structure.registry.remove(MembraneOrientationRepresentationProvider);
            this.ctx.query.structure.registry.remove(isTransmembrane);

            this.ctx.genericRepresentationControls.delete(Tag.Representation);
            this.ctx.builders.structure.representation.unregisterPreset(MembraneOrientationPreset);
        }
    },
    params: () => ({
        autoAttach: PD.Boolean(false)
    })
});

//

export const isTransmembrane = StructureSelectionQuery('Residues Embedded in Membrane', MS.struct.modifier.union([
    MS.struct.modifier.wholeResidues([
        MS.struct.modifier.union([
            MS.struct.generator.atomGroups({
                'chain-test': MS.core.rel.eq([MS.ammp('objectPrimitive'), 'atomistic']),
                'atom-test': MembraneOrientation.symbols.isTransmembrane.symbol(),
            })
        ])
    ])
]), {
    description: 'Select residues that are embedded between the membrane layers.',
    category: StructureSelectionCategory.Residue,
    ensureCustomProperties: (ctx, structure) => {
        return MembraneOrientationProvider.attach(ctx, structure);
    }
});

// //////////////////////////// TMDET VIEWER FUNCTIONS

export let membraneOrientation: MembraneOrientation;
export let pdbtmDescriptor: PDBTMDescriptor;

export async function loadWithUNITMPMembraneRepresentation(plugin: PluginUIContext, params: any) {
    // const pdbtmDescriptor: any = await downloadRegionDescriptor(plugin, params);
    pdbtmDescriptor = await downloadRegionDescriptor(plugin, params);

    membraneOrientation = createMembraneOrientation(pdbtmDescriptor);

    // load structure
    await loadStructure(plugin, params, pdbtmDescriptor);

    await createStructureRepresentation(plugin, pdbtmDescriptor);

    //
    // reset the camera because the membranes render 1st and the structure might not be fully visible
    //
    requestAnimationFrame(() => plugin.canvas3d?.requestCameraReset());
}

async function createStructureRepresentation(plugin: PluginUIContext, pdbtmDescriptor: any) {
    // get the first structure of the first model
    const structure: StateObjectRef<PMS> = plugin.managers.structure.hierarchy.current.models[0].structures[0].cell;
    const components = await createStructureComponents(plugin, structure);

    await applyTransformations(plugin, pdbtmDescriptor);

    await buildStructureRepresentation(plugin, components);

    // TODO: colors of not curated sites
    const siteColors = [
        [255,100,100], // Side1
        [100,100,255], // Side2
        [255,255,0],   // TM alpha
        [255,255,0],   // TM beta
        [255,127,0],   // TM re-entrant loop
        [0,255,  0],   // Interfacial Helix
        [196,196,196], // Unknow localization
        [0,255,  0],   // Membrane Inside
    ];
    // const siteDefinitions = pdbtmDescriptor.sites.map(
    //     (siteDefinition: any) => siteDefinition.label
    // );

    pdbtmDescriptor.chains.forEach((chain: any) => {

        let regionsBySite: PDBTMRegion[] = siteColors.map((color):PDBTMRegion => {
            // TODO: set site appropriately
            return { site: "valami", auth_ids: [], color: color };
        });

        for (let residueItem of chain.residues) {
            const siteIndex = residueItem.site_data[0].site_id_ref - 1;
            regionsBySite[siteIndex].auth_ids.push(Number(residueItem.pdb_res_label));
        }

        //console.log(`REGIONS-${chain.chain_label}`, regionsBySite);
        regionsBySite.forEach((region: PDBTMRegion) => {
            const regionUpdates = plugin.build();
            createRegionRepresentation(plugin, chain.chain_label, region, regionUpdates.to(structure));
            regionUpdates.commit();
        });

    });
}

async function createStructureComponents(plugin: PluginUIContext, structure: StateObjectCell<PMS, StateTransform<StateTransformer<StateObject<any, StateObject.Type<any>>, StateObject<any, StateObject.Type<any>>, any>>>) {
    return {
        polymer: await plugin.builders.structure.tryCreateComponentStatic(structure, 'polymer'),
        ligand: await plugin.builders.structure.tryCreateComponentStatic(structure, 'ligand'),
        water: await plugin.builders.structure.tryCreateComponentStatic(structure, 'water'),
    };
}

function createRegionRepresentation(plugin: PluginUIContext, chain: string, residue: PDBTMRegion, update: StateBuilder.To<any, any>) {
    const lastIdIndex: number = residue.auth_ids.length - 1;
    const regionLabel: string = `${chain}: ${residue.auth_ids[0]}-${residue.auth_ids[lastIdIndex]} | ${residue.site}`;
    const color: Color = Color.fromArray(residue.color, 0);
    const query: Expression = getAtomGroupExpression(chain, residue.auth_ids);

    // based on https://github.com/molstar/molstar/issues/209
    update
        .apply(StateTransforms.Model.StructureSelectionFromExpression, { label: regionLabel, expression: query })
        .apply(StateTransforms.Representation.StructureRepresentation3D, createStructureRepresentationParams(plugin, update.selector.data, {
            color: 'uniform',
            colorParams: { value: color }
        }));
}

async function buildStructureRepresentation(plugin: PluginUIContext, components: { polymer: StructureComponentType; ligand: StructureComponentType; water: StructureComponentType }) {
    const builder = plugin.builders.structure.representation;
    const update = plugin.build();
    if (components.polymer)
        builder.buildRepresentation(update, components.polymer, { type: 'cartoon', typeParams: { alpha: 0.5 } }, { tag: 'polymer' });
    if (components.ligand)
        builder.buildRepresentation(update, components.ligand, { type: 'ball-and-stick' }, { tag: 'ligand' });
    if (components.water)
        builder.buildRepresentation(update, components.water, { type: 'ball-and-stick', typeParams: { alpha: 0.6 } }, { tag: 'water' });
    await update.commit();
}

async function loadStructure(ctx: PluginUIContext, params: any, pdbtmDescriptor: PDBTMDescriptor): Promise<void> {
    const builders = ctx.builders;
    const data = await builders.data.download({
        url: params.structureUrl,
        label: `${pdbtmDescriptor.pdb_id}`,
        isBinary: false
    }); // , { state: { isGhost: true } });
    const trajectory = await builders.structure.parseTrajectory(data, 'mmcif');

    // TODO: DEBUG logs
    // console.log('ATOMIC CONFORMATION', trajectory.data!.representative.atomicConformation);
    // console.log('ATOMIC RANGES', trajectory.data!.representative.atomicRanges);
    // console.log('ATOMIC HIER', trajectory.data!.representative.atomicHierarchy);
    // console.log('ATOMIC ENTITIES', trajectory.data!.representative.entities);

    //ctx.managers.structure.hierarchy.getStructuresWithSelection();

    // create membrane representation
    await builders.structure.hierarchy.applyPreset(
        trajectory, 'default', { representationPreset: 'preset-membrane-orientation' as any });
}

async function downloadRegionDescriptor(plugin: PluginUIContext, params: any): Promise<any> {
    // run a fetch task
    const downloadResult: string = await plugin.runTask(plugin.fetch({ url: params.regionDescriptorUrl })) as string;
    const pdbtmDescriptor: any = JSON.parse(downloadResult);
    return pdbtmDescriptor;
}

// //////////////////////////// END OF TMDET VIEWER SECTION

//

export { MembraneOrientation3D };

type MembraneOrientation3D = typeof MembraneOrientation3D
const MembraneOrientation3D = PluginStateTransform.BuiltIn({
    name: 'membrane-orientation-3d',
    display: {
        name: 'Membrane Orientation',
        description: 'Membrane Orientation planes and rims. Data calculated with TMDET algorithm.'
    },
    from: PluginStateObject.Molecule.Structure,
    to: PluginStateObject.Shape.Representation3D,
    params: (a) => {
        return {
            ...MembraneOrientationParams,
        };
    }
})({
    canAutoUpdate({ oldParams, newParams }) {
        return true;
    },
    apply({ a, params }, plugin: PluginContext) {
        return Task.create('Membrane Orientation', async ctx => {
            await MembraneOrientationProvider.attach({ runtime: ctx, assetManager: plugin.managers.asset }, a.data);
            const repr = MembraneOrientationRepresentation({ webgl: plugin.canvas3d?.webgl, ...plugin.representation.structure.themes }, () => MembraneOrientationParams);
            await repr.createOrUpdate(params, a.data).runInContext(ctx);
            return new PluginStateObject.Shape.Representation3D({ repr, sourceData: a.data }, { label: 'Membrane Orientation' });
        });
    },
    update({ a, b, newParams }, plugin: PluginContext) {
        return Task.create('Membrane Orientation', async ctx => {
            await MembraneOrientationProvider.attach({ runtime: ctx, assetManager: plugin.managers.asset }, a.data);
            const props = { ...b.data.repr.props, ...newParams };
            await b.data.repr.createOrUpdate(props, a.data).runInContext(ctx);
            b.data.sourceData = a.data;
            return StateTransformer.UpdateResult.Updated;
        });
    },
    isApplicable(a) {
        return MembraneOrientationProvider.isApplicable(a.data);
    }
});

export const MembraneOrientationPreset = StructureRepresentationPresetProvider({
    id: 'preset-membrane-orientation',
    display: {
        name: 'Membrane Orientation', group: 'Annotation',
        description: 'Shows orientation of membrane layers. Data calculated with TMDET algorithm.' // TODO add ' or obtained via RCSB PDB'
    },
    isApplicable(a) {
        return MembraneOrientationProvider.isApplicable(a.data);
    },
    params: () => StructureRepresentationPresetProvider.CommonParams,
    async apply(ref, params, plugin) {
        const structureCell = StateObjectRef.resolveAndCheck(plugin.state.data, ref);
        const structure  = structureCell?.obj?.data;
        if (!structureCell || !structure) return {};

        if (!MembraneOrientationProvider.get(structure).value) {
            await plugin.runTask(Task.create('Membrane Orientation', async runtime => {
                await MembraneOrientationProvider.attach({ runtime, assetManager: plugin.managers.asset }, structure);
            }));
        }

        const membraneOrientation = await tryCreateMembraneOrientation(plugin, structureCell);
        const colorTheme = HydrophobicityColorThemeProvider.name as any;
        const preset = await PresetStructureRepresentations.auto.apply(ref, { ...params, theme: { globalName: colorTheme, focus: { name: colorTheme } } }, plugin);

        return { components: preset.components, representations: { ...preset.representations, membraneOrientation } };
    }
});

export function tryCreateMembraneOrientation(plugin: PluginContext, structure: StateObjectRef<PMS>, params?: StateTransformer.Params<MembraneOrientation3D>, initialState?: Partial<StateTransform.State>) {
    const state = plugin.state.data;
    const membraneOrientation = state.build().to(structure)
        .applyOrUpdateTagged('membrane-orientation-3d', MembraneOrientation3D, params, { state: initialState });
    return membraneOrientation.commit({ revertOnError: true });
}
